@@ -1,5 +1,5 @@
 # rawdog: RSS aggregator without delusions of grandeur.
-# Copyright 2003, 2004 Adam Sampson <azz@us-lot.org>
+# Copyright 2003, 2004, 2005 Adam Sampson <ats@offog.org>
 #
 # rawdog is free software; you can redistribute and/or modify it
 # under the terms of that license as published by the Free Software
@@ -12,16 +12,16 @@
 # General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with rawdog; see the file COPYING. If not, write to the
-# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
-# MA 02111-1307 USA, or see http://www.gnu.org/.
+# along with rawdog; see the file COPYING. If not, write to the Free
+# Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
+# MA 02110-1301, USA, or see http://www.gnu.org/.
 
-VERSION = "2.4"
+VERSION = "2.5"
 STATE_VERSION = 2
 import feedparser, feedfinder, plugins
 from persister import Persistable, Persister
-import os, time, sha, getopt, sys, re, urlparse, cgi, socket, urllib2, calendar
-import string
+import os, time, sha, getopt, sys, re, cgi, socket, urllib2, calendar
+import string, threading
 from StringIO import StringIO
 
 def set_socket_timeout(n):
@@ -213,10 +213,40 @@ def load_file(name):
 		f.close()
 	return file_cache[name]
 
+def write_ascii(f, s, config):
+	"""Write the string s, which should only contain ASCII characters, to
+	file f; if it isn't encodable in ASCII, then print a warning message
+	and write UTF-8."""
+	try:
+		f.write(s)
+	except UnicodeEncodeError, e:
+		config.bug("Error encoding output as ASCII; UTF-8 has been written instead.\n", e)
+		f.write(s.encode("UTF-8"))
+
 def short_hash(s):
 	"""Return a human-manipulatable 'short hash' of a string."""
 	return sha.new(s).hexdigest()[-8:]
 
+def decode_structure(struct, encoding):
+	"""Walk through a structure returned by feedparser, decoding any
+	strings that haven't already been converted to Unicode."""
+	def is_dict(t):
+		return (t is dict) or (t is feedparser.FeedParserDict)
+	for (key, value) in struct.items():
+		if type(value) is str:
+			try:
+				struct[key] = value.decode(encoding)
+			except:
+				# If the encoding's invalid, at least preserve
+				# the byte stream.
+				struct[key] = value.decode("ISO-8859-1")
+		elif is_dict(type(value)):
+			decode_structure(value, encoding)
+		elif type(value) is list:
+			for item in value:
+				if is_dict(type(item)):
+					decode_structure(item, encoding)
+
 non_alphanumeric_re = re.compile(r'<[^>]*>|\&[^\;]*\;|[^a-z0-9]')
 class Feed:
 	"""An RSS feed."""
@@ -237,13 +267,9 @@ class Feed:
 			return 0
 		else:
 			return 1
-	
-	def update(self, rawdog, now, config):
-		"""Fetch articles from a feed and add them to the collection.
-		Returns True if any articles were read, False otherwise."""
 
-		articles = rawdog.articles
-		handlers = []
+	def fetch(self, rawdog, config):
+		"""Fetch the current set of articles from the feed."""
 
 		class DummyPasswordMgr:
 			def __init__(self, creds):
@@ -253,15 +279,17 @@ class Feed:
 			def find_user_password(self, realm, authuri):
 				return self.creds
 
+		handlers = []
+
 		if self.args.has_key("user") and self.args.has_key("password"):
 			mgr = DummyPasswordMgr((self.args["user"], self.args["password"]))
 			handlers.append(urllib2.HTTPBasicAuthHandler(mgr))
 
 		proxies = {}
-		for key in self.args.keys():
+		for key, arg in self.args.items():
 			if key.endswith("_proxy"):
-				proxies[key[:-6]] = self.args[key]
-		if len(proxies.keys()) != 0:
+				proxies[key[:-6]] = arg
+		if len(proxies) != 0:
 			handlers.append(urllib2.ProxyHandler(proxies))
 
 		if self.args.has_key("proxyuser") and self.args.has_key("proxypassword"):
@@ -270,26 +298,29 @@ class Feed:
 
 		plugins.call_hook("add_urllib2_handlers", rawdog, config, self, handlers)
 
-		feedparser._FeedParserMixin.can_contain_relative_uris = ["url"]
-		feedparser._FeedParserMixin.can_contain_dangerous_markup = []
 		try:
-			p = feedparser.parse(self.url,
+			return feedparser.parse(self.url,
 				etag = self.etag,
 				modified = self.modified,
 				agent = "rawdog/" + VERSION,
 				handlers = handlers)
-			status = p.get("status")
 		except:
-			p = None
-			status = None
+			return None
 
+	def update(self, rawdog, now, config, p):
+		"""Add new articles from a feed to the collection.
+		Returns True if any articles were read, False otherwise."""
+
+		status = None
+		if p is not None:
+			status = p.get("status")
 		self.last_update = now
 
 		error = None
 		non_fatal = False
 		old_url = self.url
 		if p is None:
-			error = "Error parsing feed."
+			error = "Error fetching or parsing feed."
 		elif status is None and len(p["feed"]) == 0:
 			if config["ignoretimeouts"]:
 				return False
@@ -329,6 +360,8 @@ class Feed:
 			if not non_fatal:
 				return False
 
+		decode_structure(p, p.get("encoding") or "UTF-8")
+
 		self.etag = p.get("etag")
 		self.modified = p.get("modified")
 
@@ -340,6 +373,7 @@ class Feed:
 
 		self.feed_info = p["feed"]
 		feed = self.url
+		articles = rawdog.articles
 
 		seen = {}
 		sequence = 0
@@ -488,9 +522,9 @@ def parse_time(value, default = "m"):
 	the default argument to change this. Raises ValueError if the format
 	isn't recognised."""
 	units = { "s" : 1, "m" : 60, "h" : 3600, "d" : 86400, "w" : 604800 }
-	for unit in units.keys():
+	for unit, size in units.items():
 		if value.endswith(unit):
-			return int(value[:-len(unit)]) * units[unit]
+			return int(value[:-len(unit)]) * size
 	return int(value) * units[default]
 
 def parse_bool(value):
@@ -530,6 +564,7 @@ class Config:
 
 	def __init__(self):
 		self.files_loaded = []
+		self.loglock = threading.Lock()
 		self.reset()
 
 	def reset(self):
@@ -560,6 +595,7 @@ class Config:
 			"hideduplicates" : "",
 			"newfeedperiod" : "3h",
 			"changeconfig": 0,
+			"numthreads": 0,
 			}
 
 	def __getitem__(self, key): return self.config[key]
@@ -670,6 +706,8 @@ class Config:
 			self["newfeedperiod"] = l[1]
 		elif l[0] == "changeconfig":
 			self["changeconfig"] = parse_bool(l[1])
+		elif l[0] == "numthreads":
+			self["numthreads"] = int(l[1])
 		elif l[0] == "include":
 			self.load(l[1], False)
 		elif plugins.call_hook("config_option_arglines", self, l[0], l[1], arglines):
@@ -685,7 +723,9 @@ class Config:
 	def log(self, *args):
 		"""If running in verbose mode, print a status message."""
 		if self["verbose"]:
+			self.loglock.acquire()
 			print >>sys.stderr, "".join(map(str, args))
+			self.loglock.release()
 
 	def bug(self, *args):
 		"""Report detection of a bug in rawdog."""
@@ -738,14 +778,78 @@ class ChangeFeedEditor:
 				line = line.replace(self.oldurl, self.newurl, 1)
 			outputfile.write(line)
 
+class FeedFetcher:
+	"""Class that will handle fetching a set of feeds in parallel."""
+
+	def __init__(self, rawdog, feedlist, config):
+		self.rawdog = rawdog
+		self.config = config
+		self.lock = threading.Lock()
+		self.jobs = {}
+		for feed in feedlist:
+			self.jobs[feed] = 1
+		self.results = {}
+
+	def worker(self, num):
+		rawdog = self.rawdog
+		config = self.config
+
+		config.log("Thread ", num, " starting")
+		while 1:
+			self.lock.acquire()
+			if self.jobs == {}:
+				job = None
+			else:
+				job = self.jobs.keys()[0]
+				del self.jobs[job]
+			self.lock.release()
+			if job is None:
+				break
+
+			config.log("Thread ", num, " fetching feed: ", job)
+			feed = rawdog.feeds[job]
+			plugins.call_hook("pre_update_feed", rawdog, config, feed)
+			self.results[job] = feed.fetch(rawdog, config)
+		config.log("Thread ", num, " done")
+
+	def run(self, numworkers):
+		self.config.log("Thread farm starting with ", len(self.jobs), " jobs")
+		workers = []
+		for i in range(numworkers):
+			self.lock.acquire()
+			isempty = (self.jobs == {})
+			self.lock.release()
+			if isempty:
+				# No jobs left in the queue -- don't bother
+				# starting any more workers.
+				break
+
+			t = threading.Thread(target = self.worker, args = (i,))
+			t.start()
+			workers.append(t)
+		for worker in workers:
+			worker.join()
+		self.config.log("Thread farm finished with ", len(self.results), " results")
+		return self.results
+
 class Rawdog(Persistable):
 	"""The aggregator itself."""
 
 	def __init__(self):
 		self.feeds = {}
 		self.articles = {}
+		self.plugin_storage = {}
 		self.state_version = STATE_VERSION
 
+	def get_plugin_storage(self, plugin):
+		try:
+			st = self.plugin_storage.setdefault(plugin, {})
+		except AttributeError:
+			# rawdog before 2.5 didn't have plugin storage.
+			st = {}
+			self.plugin_storage = {plugin: st}
+		return st
+
 	def check_state_version(self):
 		"""Check the version of the state file."""
 		try:
@@ -778,8 +882,8 @@ class Rawdog(Persistable):
 		print >>sys.stderr, "Feed URL automatically changed."
 
 	def list(self, config):
-		for url in self.feeds.keys():
-			feed = self.feeds[url]
+		"""List the configured feeds."""
+		for url, feed in self.feeds.items():
 			feed_info = feed.feed_info
 			print url
 			print "  ID:", feed.get_id(config)
@@ -788,6 +892,8 @@ class Rawdog(Persistable):
 			print "  Link:", feed_info.get("link")
 
 	def sync_from_config(self, config):
+		"""Update rawdog's internal state to match the
+		configuration."""
 		seenfeeds = {}
 		for (url, period, args) in config["feedslist"]:
 			seenfeeds[url] = 1
@@ -817,9 +923,13 @@ class Rawdog(Persistable):
 				self.modified()
 
 	def update(self, config, feedurl = None):
+		"""Perform the update action: check feeds for new articles, and
+		expire old ones."""
 		config.log("Starting update")
 		now = time.time()
 
+		feedparser._FeedParserMixin.can_contain_relative_uris = ["url"]
+		feedparser._FeedParserMixin.can_contain_dangerous_markup = []
 		set_socket_timeout(config["timeout"])
 
 		if feedurl is None:
@@ -836,14 +946,25 @@ class Rawdog(Persistable):
 		numfeeds = len(update_feeds)
 		config.log("Will update ", numfeeds, " feeds")
 
+		if config["numthreads"] > 0:
+			fetcher = FeedFetcher(self, update_feeds, config)
+			prefetched = fetcher.run(config["numthreads"])
+		else:
+			prefetched = {}
+
 		count = 0
 		seen_some_items = {}
 		for url in update_feeds:
 			count += 1
 			config.log("Updating feed ", count, " of " , numfeeds, ": ", url)
 			feed = self.feeds[url]
-			plugins.call_hook("pre_update_feed", self, config, feed)
-			rc = feed.update(self, now, config)
+			if url in prefetched:
+				content = prefetched[url]
+			else:
+				plugins.call_hook("pre_update_feed", self, config, feed)
+				content = feed.fetch(self, config)
+			plugins.call_hook("mid_update_feed", self, config, feed, content)
+			rc = feed.update(self, now, config, content)
 			plugins.call_hook("post_update_feed", self, config, feed, rc)
 			if rc:
 				seen_some_items[url] = 1
@@ -855,12 +976,13 @@ class Rawdog(Persistable):
 				plugins.call_hook("article_expired", self, config, article, now)
 				count += 1
 				del self.articles[key]
-		config.log("Expired ", count, " articles, leaving ", len(self.articles.keys()))
+		config.log("Expired ", count, " articles, leaving ", len(self.articles))
 
 		self.modified()
 		config.log("Finished update")
 
 	def get_template(self, config):
+		"""Get the main template."""
 		if config["template"] != "default":
 			return load_file(config["template"])
 
@@ -895,13 +1017,15 @@ __feeds__
 <p id="aboutrawdog">Generated by
 <a href="http://offog.org/code/rawdog.html">rawdog</a>
 version __version__
-by <a href="mailto:azz@us-lot.org">Adam Sampson</a>.</p>
+by <a href="mailto:ats@offog.org">Adam Sampson</a>.</p>
 </div>
 </body>
-</html>"""
+</html>
+"""
 		return template
 
 	def get_itemtemplate(self, config):
+		"""Get the item template."""
 		if config["itemtemplate"] != "default":
 			return load_file(config["itemtemplate"])
 
@@ -919,16 +1043,140 @@ __description__
 		return template
 
 	def show_template(self, config):
+		"""Show the configured main template."""
 		print self.get_template(config)
 
 	def show_itemtemplate(self, config):
+		"""Show the configured item template."""
 		print self.get_itemtemplate(config)
 
-	def write(self, config):
-		outputfile = config["outputfile"]
-		config.log("Starting write")
-		now = time.time()
+	def write_article(self, f, article, config):
+		"""Write an article to the given file."""
+		feed = self.feeds[article.feed]
+		feed_info = feed.feed_info
+		entry_info = article.entry_info
 
+		link = entry_info.get("link")
+		if link == "":
+			link = None
+
+		guid = entry_info.get("id")
+		if guid == "":
+			guid = None
+
+		itembits = {}
+		for name, value in feed.args.items():
+			if name.startswith("define_"):
+				itembits[name[7:]] = value
+
+		title = detail_to_html(entry_info.get("title_detail"), True, config)
+
+		key = None
+		for k in ["content", "summary_detail"]:
+			if entry_info.has_key(k):
+				key = k
+				break
+		if key is None:
+			description = None
+		else:
+			force_preformatted = feed.args.has_key("format") and (feed.args["format"] == "text")
+			description = detail_to_html(entry_info[key], False, config, force_preformatted)
+
+		date = article.date
+		if title is None:
+			if link is None:
+				title = "Article"
+			else:
+				title = "Link"
+
+		itembits["title_no_link"] = title
+		if link is not None:
+			itembits["url"] = string_to_html(link, config)
+		else:
+			itembits["url"] = ""
+		if guid is not None:
+			itembits["guid"] = string_to_html(guid, config)
+		else:
+			itembits["guid"] = ""
+		if link is None:
+			itembits["title"] = title
+		else:
+			itembits["title"] = '<a href="' + string_to_html(link, config) + '">' + title + '</a>'
+
+		itembits["feed_title_no_link"] = detail_to_html(feed_info.get("title_detail"), True, config)
+		itembits["feed_title"] = feed.get_html_link(config)
+		itembits["feed_url"] = string_to_html(feed.url, config)
+		itembits["feed_hash"] = short_hash(feed.url)
+		itembits["feed_id"] = feed.get_id(config)
+		itembits["hash"] = short_hash(article.hash)
+
+		if description is not None:
+			itembits["description"] = description
+		else:
+			itembits["description"] = ""
+
+		author = author_to_html(entry_info, feed.url, config)
+		if author is not None:
+			itembits["author"] = author
+		else:
+			itembits["author"] = ""
+
+		itembits["added"] = format_time(article.added, config)
+		if date is not None:
+			itembits["date"] = format_time(date, config)
+		else:
+			itembits["date"] = ""
+
+		plugins.call_hook("output_item_bits", self, config, feed, article, itembits)
+		itemtemplate = self.get_itemtemplate(config)
+		f.write(fill_template(itemtemplate, itembits))
+
+	def write_remove_dups(self, articles, config, now):
+		"""Filter the list of articles to remove articles that are too
+		old or are duplicates."""
+		kept_articles = []
+		seen_links = {}
+		seen_guids = {}
+		dup_count = 0
+		for article in articles:
+			age = now - article.added
+			if config["maxage"] != 0 and age > config["maxage"]:
+				continue
+
+			feed = self.feeds[article.feed]
+			entry_info = article.entry_info
+
+			link = entry_info.get("link")
+			if link == "":
+				link = None
+
+			guid = entry_info.get("id")
+			if guid == "":
+				guid = None
+
+			if feed.args.get("allowduplicates") != "true":
+				is_dup = False
+				for key in config["hideduplicates"]:
+					if key == "id" and guid is not None:
+						if seen_guids.has_key(guid):
+							is_dup = True
+						seen_guids[guid] = 1
+						break
+					elif key == "link" and link is not None:
+						if seen_links.has_key(link):
+							is_dup = True
+						seen_links[link] = 1
+						break
+				if is_dup:
+					dup_count += 1
+					continue
+
+			kept_articles.append(article)
+		return (kept_articles, dup_count)
+
+	def get_main_template_bits(self, config):
+		"""Get the bits that are used in the default main template,
+		with the exception of items and num_items."""
 		bits = { "version" : VERSION }
 		bits.update(config["defines"])
 
@@ -938,6 +1186,62 @@ __description__
 
 		bits["refresh"] = """<meta http-equiv="Refresh" """ + 'content="' + str(refresh) + '"' + """>"""
 
+		f = StringIO()
+		print >>f, """<table id="feeds">
+<tr id="feedsheader">
+<th>Feed</th><th>RSS</th><th>Last update</th><th>Next update</th>
+</tr>"""
+		feeds = self.feeds.values()
+		feeds.sort(lambda a, b: cmp(a.get_html_name(config).lower(), b.get_html_name(config).lower()))
+		for feed in feeds:
+			print >>f, '<tr class="feedsrow">'
+			print >>f, '<td>' + feed.get_html_link(config) + '</td>'
+			print >>f, '<td><a class="xmlbutton" href="' + cgi.escape(feed.url) + '">XML</a></td>'
+			print >>f, '<td>' + format_time(feed.last_update, config) + '</td>'
+			print >>f, '<td>' + format_time(feed.last_update + feed.period, config) + '</td>'
+			print >>f, '</tr>'
+		print >>f, """</table>"""
+		bits["feeds"] = f.getvalue()
+		bits["num_feeds"] = str(len(feeds))
+
+		return bits
+
+	def write_output_file(self, articles, article_dates, config):
+		"""Write a regular rawdog HTML output file."""
+		f = StringIO()
+		dw = DayWriter(f, config)
+		plugins.call_hook("output_items_begin", self, config, f)
+
+		for article in articles:
+			if not plugins.call_hook("output_items_heading", self, config, f, article, article_dates[article]):
+				dw.time(article_dates[article])
+
+			self.write_article(f, article, config)
+
+		dw.close()
+		plugins.call_hook("output_items_end", self, config, f)
+
+		bits = self.get_main_template_bits(config)
+		bits["items"] = f.getvalue()
+		bits["num_items"] = str(len(self.articles))
+		plugins.call_hook("output_bits", self, config, bits)
+		s = fill_template(self.get_template(config), bits)
+		outputfile = config["outputfile"]
+		if outputfile == "-":
+			write_ascii(sys.stdout, s, config)
+		else:
+			config.log("Writing output file: ", outputfile)
+			f = open(outputfile + ".new", "w")
+			write_ascii(f, s, config)
+			f.close()
+			os.rename(outputfile + ".new", outputfile)
+
+	def write(self, config):
+		"""Perform the write action: write articles to the output
+		file."""
+		config.log("Starting write")
+		now = time.time()
+
 		article_dates = {}
 		articles = self.articles.values()
 		for a in articles:
@@ -970,159 +1274,15 @@ __description__
 
 		plugins.call_hook("output_write", self, config, articles)
 
-		f = StringIO()
-		itemtemplate = self.get_itemtemplate(config)
-		dw = DayWriter(f, config)
-		plugins.call_hook("output_items_begin", self, config, f)
-
-		seen_links = {}
-		seen_guids = {}
-
-		count = 0
-		dup_count = 0
-		for article in articles:
-			age = now - article.added
-			if config["maxage"] != 0 and age > config["maxage"]:
-				break
-
-			feed = self.feeds[article.feed]
-			feed_info = feed.feed_info
-			entry_info = article.entry_info
-
-			link = entry_info.get("link")
-			if link == "":
-				link = None
-
-			guid = entry_info.get("id")
-			if guid == "":
-				guid = None
-
-			if feed.args.get("allowduplicates") != "true":
-				is_dup = False
-				for key in config["hideduplicates"]:
-					if key == "id" and guid is not None:
-						if seen_guids.has_key(guid):
-							is_dup = True
-						seen_guids[guid] = 1
-						break
-					elif key == "link" and link is not None:
-						if seen_links.has_key(link):
-							is_dup = True
-						seen_links[link] = 1
-						break
-				if is_dup:
-					dup_count += 1
-					continue
-
-			count += 1
-			if not plugins.call_hook("output_items_heading", self, config, f, article, article_dates[article]):
-				dw.time(article_dates[article])
-
-			itembits = {}
-			for name, value in feed.args.items():
-				if name.startswith("define_"):
-					itembits[name[7:]] = value
-
-			title = detail_to_html(entry_info.get("title_detail"), True, config)
-
-			key = None
-			for k in ["content", "summary_detail"]:
-				if entry_info.has_key(k):
-					key = k
-					break
-			if key is None:
-				description = None
-			else:
-				force_preformatted = feed.args.has_key("format") and (feed.args["format"] == "text")
-				description = detail_to_html(entry_info[key], False, config, force_preformatted)
-
-			date = article.date
-			if title is None:
-				if link is None:
-					title = "Article"
-				else:
-					title = "Link"
-
-			itembits["title_no_link"] = title
-			if link is not None:
-				itembits["url"] = string_to_html(link, config)
-			else:
-				itembits["url"] = ""
-			if guid is not None:
-				itembits["guid"] = string_to_html(guid, config)
-			else:
-				itembits["guid"] = ""
-			if link is None:
-				itembits["title"] = title
-			else:
-				itembits["title"] = '<a href="' + string_to_html(link, config) + '">' + title + '</a>'
-
-			itembits["feed_title_no_link"] = detail_to_html(feed_info.get("title_detail"), True, config)
-			itembits["feed_title"] = feed.get_html_link(config)
-			itembits["feed_url"] = string_to_html(feed.url, config)
-			itembits["feed_hash"] = short_hash(feed.url)
-			itembits["feed_id"] = feed.get_id(config)
-			itembits["hash"] = short_hash(article.hash)
-
-			if description is not None:
-				itembits["description"] = description
-			else:
-				itembits["description"] = ""
-
-			author = author_to_html(entry_info, feed.url, config)
-			if author is not None:
-				itembits["author"] = author
-			else:
-				itembits["author"] = ""
-
-			itembits["added"] = format_time(article.added, config)
-			if date is not None:
-				itembits["date"] = format_time(date, config)
-			else:
-				itembits["date"] = ""
-
-			plugins.call_hook("output_item_bits", self, config, feed, article, itembits)
-			f.write(fill_template(itemtemplate, itembits))
-
-		dw.close()
-		plugins.call_hook("output_items_end", self, config, f)
-
-		bits["items"] = f.getvalue()
-		bits["num_items"] = str(numarticles)
-		config.log("Selected ", count, " of ", numarticles, " articles to write; ignored ", dup_count, " duplicates")
-
-		f = StringIO()
-		print >>f, """<table id="feeds">
-<tr id="feedsheader">
-<th>Feed</th><th>RSS</th><th>Last update</th><th>Next update</th>
-</tr>"""
-		feeds = self.feeds.values()
-		feeds.sort(lambda a, b: cmp(a.get_html_name(config).lower(), b.get_html_name(config).lower()))
-		for feed in feeds:
-			print >>f, '<tr class="feedsrow">'
-			print >>f, '<td>' + feed.get_html_link(config) + '</td>'
-			print >>f, '<td><a class="xmlbutton" href="' + cgi.escape(feed.url) + '">XML</a></td>'
-			print >>f, '<td>' + format_time(feed.last_update, config) + '</td>'
-			print >>f, '<td>' + format_time(feed.last_update + feed.period, config) + '</td>'
-			print >>f, '</tr>'
-		print >>f, """</table>"""
-		bits["feeds"] = f.getvalue()
-		bits["num_feeds"] = str(len(feeds))
-
-		plugins.call_hook("output_bits", self, config, bits)
-		s = fill_template(self.get_template(config), bits)
-		if outputfile == "-":
-			print s
+		if not plugins.call_hook("output_sorted_filter", self, config, articles):
+			(articles, dup_count) = self.write_remove_dups(articles, config, now)
 		else:
-			config.log("Writing output file: ", outputfile)
-			f = open(outputfile + ".new", "w")
-			try:
-				print >>f, s
-			except UnicodeEncodeError, e:
-				config.bug("Error encoding output as ASCII; UTF-8 has been written instead.\n", e)
-				print >>f, s.encode("UTF-8")
-			f.close()
-			os.rename(outputfile + ".new", outputfile)
+			dup_count = 0
+
+		config.log("Selected ", len(articles), " of ", numarticles, " articles to write; ignored ", dup_count, " duplicates")
+
+		if not plugins.call_hook("output_write_files", self, config, articles, article_dates):
+			self.write_output_file(articles, article_dates, config)
 
 		config.log("Finished write")
 
@@ -1134,6 +1294,7 @@ Usage: rawdog [OPTION]...
 General options (use only once):
 -d|--dir DIR                 Use DIR instead of ~/.rawdog
 -v, --verbose                Print more detailed status information
+-N, --no-locking             Do not lock the state file
 --help                       Display this help and exit
 
 Actions (performed in order given):
@@ -1151,13 +1312,13 @@ Special actions (all other options are ignored if one of these is specified):
 --upgrade OLDDIR NEWDIR      Import feed state from rawdog 1.x directory
                              OLDDIR into rawdog 2.x directory NEWDIR
 
-Report bugs to <azz@us-lot.org>."""
+Report bugs to <ats@offog.org>."""
 
 def main(argv):
 	"""The command-line interface to the aggregator."""
 
 	try:
-		(optlist, args) = getopt.getopt(argv, "ulwf:c:tTd:va:", ["update", "list", "write", "update-feed=", "help", "config=", "show-template", "dir=", "show-itemtemplate", "verbose", "upgrade", "add="])
+		(optlist, args) = getopt.getopt(argv, "ulwf:c:tTd:va:N", ["update", "list", "write", "update-feed=", "help", "config=", "show-template", "dir=", "show-itemtemplate", "verbose", "upgrade", "add=", "no-locking"])
 	except getopt.GetoptError, s:
 		print s
 		usage()
@@ -1174,6 +1335,7 @@ def main(argv):
 
 	statedir = os.environ["HOME"] + "/.rawdog"
 	verbose = 0
+	locking = 1
 	for o, a in optlist:
 		if o == "--help":
 			usage()
@@ -1182,6 +1344,8 @@ def main(argv):
 			statedir = a
 		elif o in ("-v", "--verbose"):
 			verbose = 1
+		elif o in ("-N", "--no-locking"):
+			locking = 0
 
 	try:
 		os.chdir(statedir)
@@ -1199,7 +1363,7 @@ def main(argv):
 	if verbose:
 		config["verbose"] = True
 
-	persister = Persister("state", Rawdog)
+	persister = Persister("state", Rawdog, locking)
 	try:
 		rawdog = persister.load()
 	except KeyboardInterrupt:
